import sys
import wave

if len(sys.argv) != 4:
    print("Usage: python3 wav_to_5_bit_lookup.py input_file [a/r] [s/c]")
    print("'a' means Arduino type data structure, 'r' means text-table to load into R for checking")
    print("'s' means segmenting the file by silence, 'c' means convert complete file.")
    print("writes a bunch of files:")
    print("- sounds.txt  -  the 5-bit sound data file")
    print("- realsounds.txt  - the original 8 bit sound as data table for checking")
    print("- startstop.txt - start and stop indices for the arduino")
    exit()

wavfile =  sys.argv[1]
toarduino = sys.argv[2] # a=arduino, r=r
segment_or_complete = sys.argv[3]
w = wave.open(wavfile)
bit5_wave = []
silence_segment = [] ## memorize the silent 8-byte segments for cutting out the frames!

# removes silence ranges that mark the gaps between the instruments (gaps longer than 30ms)
# uses the silence ranges to define the start-points and stop points of the instruments

numbytes = w.getnframes()
outfile = "sounds.txt"
posfile = "startstop.txt"
realfile = "realsounds.txt"

forward_lookup= [0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,2,2,2,3,3,3,3,4,4,4,5,5,5,6,6,7,7,8,8,9,9,10,11,11,12,12,13,14,14,15,16,17,17,18,19,19,20,20,21,22,22,23,23,24,24,25,25,26,26,26,27,27,27,28,28,28,28,29,29,29,29,29,30,30,30,30,30,30,30,30,30,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31]


def five_highest_bits(frame):
    bits = []
    for b in frame:
        corr_byte = forward_lookup[(b >> 1)]
        for i in [4,3,2,1,0]:
            bits.append((corr_byte >> i) & 1)
    return(bits)

def bitarray_to_byte(bitarray):
    byte = 0
    a = 128
    for b in range(8):
        byte = byte + a*bitarray[b]
        a = a >> 1
    return(byte)

# converting the whole thing into 5-bit wav
i = 0
realf = open(realfile, "w")
while i < (numbytes - 7):
    frame = w.readframes(8)
    for b in frame:
        print(b, file=realf)
    # checking whether all bytes in the current frame are silence:
    is_silence = True # true when not testing script
    if segment_or_complete == "c":
        is_silence = False
    for a in frame:
        if not (a >= 120 and a <= 135): 
            is_silence = False
    i += 8
    bitseq = five_highest_bits(frame)
    for a in range(5):
        bit5_wave.append(str(int(bitarray_to_byte(bitseq[(a*8):(a*8)+8]))))
    silence_segment.append(is_silence)

print("total length: ", len(bit5_wave), " bytes.")
realf.close()


# removing long silence segments between the sounds
maxzeros = 50 # 50 5-byte blocks are marked with zero
blockopen = False
bit5_wave2 = [] ## an array of arrays
silenceblock = ["123", "222", "247", "189", "239"] # codes for 8 x 15
zeros = 0
i = 0
for i in range(len(silence_segment)):
    data_segment = bit5_wave[(i*5):((i*5)+5)] 
    if silence_segment[i]: # silence
        zeros += 1 
    else: # actual data
        if not blockopen:
            blockopen = True
            temp = data_segment
        else:
            if zeros > 0:
                if zeros < maxzeros:
                    temp = temp + silenceblock * zeros
                    temp = temp + data_segment
                else:
                    blockopen = False
                    bit5_wave2.append(temp)
            else:
                temp = temp + data_segment
        zeros = 0

if blockopen:
   bit5_wave2.append(temp) 

bit5_wave_out = silenceblock
starts = []
stops = []

for i in range(len(bit5_wave2)):
    starts.append(str(len(bit5_wave_out)))
    bit5_wave_out = bit5_wave_out + bit5_wave2[i] + silenceblock
    stops.append(str(len(bit5_wave_out)-1))

print("final length: ", len(bit5_wave_out), " bytes.")
print("segments: ", len(stops))

outf = open(outfile,"w")

if toarduino == "a":
    print("{" + ",".join(bit5_wave_out)+"}", file=outf)    

if toarduino == "r":
    print("\n".join(bit5_wave_out), file=outf)

outf.close()

outf = open(posfile,"w")
print("starts={" + ",".join(starts)+"}", file=outf)
print("stops={" + ",".join(stops)+"}", file=outf)
outf.close()
    

## R code for quality check
'''
library(pheatmap);
# loading the real waveform (lets say 7 bits)
wave = floor(as.matrix(read.delim("realsounds.txt", header=F, sep=" "))[,1] / 2)
plot(wave[1:3000], type="l")
points(wave[1:3000], pch=".", cex=2)


tobits = function(bytevect){
	bits = rep(0, 40)
	for (a in 1:5)
        	bits[((a-1)*8+1):(a*8)] = rev(as.numeric(intToBits(bytevect[a])[1:8]))
	bits
	}

# 8 x 5 bits into one byte
tobyte = function(bitvect){
        vals = rep(0,8)
        for (b in 1:8) {
		val = 0
		pos = c(16,8,4,2,1)
		for (a in 1:5)
	 		val = val + pos[a] * bitvect[(b-1) * 5 + a]
		vals[b] = val
		}
	vals
	}

rr=read.delim("sounds.txt", header=F)[,1]
out=NULL
ind = 1
while (ind <= length(rr)) {
	piece = rr[ind:(ind+4)]
	ind = ind + 5
	out = c(out, tobyte(tobits(piece)))
}

firstpart = c(0,0,0,0,8,8,8,8,8,8,8,8,8,8,16,16,16,16,16,16,16,16,16,24,24,24,24,24,24,24,24,24)
lookup = firstpart
for (a in 32:95)
  lookup[a+1] = round(round(a / 2.75) * 2.75)
lookup[97:128] = 127 - rev(firstpart)

lookup_out = unique(lookup)

out2 = NA
for (a in 1:length(out))
  out2[a] = lookup_out[out[a]+1]

points(out2[1:3000], type="l", col="blue")
points(out2[1:3000], pch=".", cex=2, col="blue")








'''



